/*
** Module   : ContainerTree.java
** Abstract : The tree Widget Browser component.
**
** Copyright (c) 2014-2015, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- ---------------------------------Description----------------------------------
** 001 HC  20141031 Created initial version
** 002 HC  20141118 Added object identity hash-code property.
** 003 HC  20141130 Added the capability to display widget config fields.
** 004 HC  20141221 New features added: property table now contains all widget and its associated
**                  config fields, added filter textbox, if property value is a weak reference it
**                  is dereferenced, property value string formatter recognizes more well-known
**                  types.
** 005 HC  20150216 Widget properties can be copied to clipboard.
*/
package com.goldencode.p2j.ui.client.widgetbrowser;

import java.awt.Toolkit;
import java.awt.datatransfer.*;
import java.awt.event.ActionListener;
import java.lang.ref.WeakReference;
import java.lang.reflect.*;
import java.util.*;

import javax.swing.JTree;
import javax.swing.event.TreeSelectionEvent;
import javax.swing.event.TreeSelectionListener;
import javax.swing.tree.*;

import org.apache.commons.lang.builder.*;

import com.goldencode.p2j.ui.*;
import com.goldencode.p2j.ui.client.*;
import com.goldencode.p2j.ui.client.widget.*;

/**
 * The tree Widget Browser component.
 */
@SuppressWarnings("serial")
public class ContainerTree 
extends JTree 
{	
	/**
	 * Full class names flag.
	 */
   private boolean showFullClassName = false;

	/**
	 * Event listener.
	 */
   private ActionListener actionListener;
	
	/**
	 * Related property table components.
	 */
   private PropertyTable table;
   
   /**
    * The property table inclusion strings.
    */
   private Iterable<String> inclusions = null;
	
	/**
	 * The root container.
	 */
   private Container<?> container;

	/**
	 * The constructor.
	 * 
	 * @param   container
	 *          The root container.
	 * @param   table
	 *          The related property table component.
	 */
   public ContainerTree(Container<?> container, final PropertyTable table) 
	{
		super();
		this.table = table;
		this.container = container;
		
		setModel(new DefaultTreeModel(sortNodes(buildTreeModel(container))));

		// disable multiple selection
		getSelectionModel().setSelectionMode(TreeSelectionModel.SINGLE_TREE_SELECTION);

		addTreeSelectionListener(new TreeSelectionListener() 
		{
			public void valueChanged(TreeSelectionEvent e) 
			{
			   refreshPropertyTable();
			}
		});
	}
	
	/**
	 * Sets the event listener.
    *
	 * @param   actionListener
	 *          Valid {@link ActionListener} reference.
	 */
   public void setActionListener(ActionListener actionListener)
   {
      this.actionListener = actionListener;
   }

   /**
    * Returns the state of show full class names flag.
    * 
    * @return  The show full class name flag.
    */
   public boolean isShowFullClassName()
	{
		return showFullClassName;
	}

   /**
    * Sets the state of show full class names flag.
    * 
    * @param   showFullClassName
    *          The show full class name flag.
    */
	public void setShowFullClassName(boolean showFullClassName) 
	{
		this.showFullClassName = showFullClassName;
	}
	
	/**
	 * Setter for the property inclusions. 
	 * 
	 * @param   inclusions
	 *          The setter value.
	 */
	public void setPropertyInclusions(Iterable<String> inclusions)
	{
	   this.inclusions = inclusions;
	}

	/**
    * Refreshes (rebuilds) the tree nodes.
    */
   public void refresh()
   {
      // store selected node
      DefaultMutableTreeNode selectedNode = (DefaultMutableTreeNode) getLastSelectedPathComponent();
   
      // store expanded widgets
      List<Widget<?>> expandedWidgets = getExpandedWidgets();
   
      // rebuild tree
      ((DefaultTreeModel) getModel()).setRoot(sortNodes(buildTreeModel(container)));
   
      // restore expanded nodes
      expandRows(findNodes(expandedWidgets));
   
      // restore selected node
      TreePath result = null;
      if (selectedNode instanceof DefaultMutableTreeNode
            && ((DefaultMutableTreeNode) selectedNode).getUserObject() instanceof WidgetWrapper)
      {
         result = findNode(((WidgetWrapper) ((DefaultMutableTreeNode) selectedNode).getUserObject())
                     .getWidget());
      }
   
      if (result == null)
      {
         // clear table data - selected node doesn't exist
         table.resetProperties();
         return;
      }
   
      setSelectionPath(result);
   }
   
   /**
    * Refreshes the property table associated with this tree component.
    */
   public void refreshPropertyTable()
   {
      DefaultMutableTreeNode node = (DefaultMutableTreeNode) getLastSelectedPathComponent();

      if (node == null || !(node.getUserObject() instanceof WidgetWrapper)) 
      {
         return;
      }
      
      table.setTableContent(((WidgetWrapper) node.getUserObject()).getProperties(inclusions));
      
      if (actionListener != null)
      {
         actionListener.actionPerformed(
               new Event(this, 
                         Event.WIDGET_SELECTION_CHANGED, 
                         ((WidgetWrapper) node.getUserObject()).getWidget()));
      }
   }

   /**
    * Collapse all tree nodes.
    */
   public void collapseAll()
   {
      for (int i = getRowCount() - 1; i >= 1; i--) 
      {
         collapseRow(i);
      }
   }

   /** 
    * Expand all tree nodes.
    */
   public void expandAll()
   {
      for (int i = 0; i < getRowCount(); i++) 
      {
         expandRow(i);
      }   
   }

   /**
    * Returns the related tree path for the supplied widget.
    * 
    * @param   subject
    *          The {@link Widget} reference.
    *          
    * @return  {@link TreePath} instance or <code>null</code> if not found.
    */
   public TreePath findNode(Widget<?> subject)
   {
      List<Widget<?>> subjects = new ArrayList<>();
      subjects.add(subject);
      List<TreePath> result = findNodes(subjects);
      return result == null || result.isEmpty() ? null : result.get(0);
   }

   /**
    * Checks whether the supplied node is selected.
    * 
    * @param   widget
    *          The {@link Widget} reference.
    *          
    * @return  <code>true</code> if selected, <code>false</code> otherwise.
    */
   public boolean isSelected(Widget<?> widget)
   {
      DefaultMutableTreeNode selectedNode = 
            (DefaultMutableTreeNode) getLastSelectedPathComponent();
      if (selectedNode == null || !(selectedNode.getUserObject() instanceof WidgetWrapper))
      {
         return false;
      }
      
      return Objects.equals(widget, ((WidgetWrapper) selectedNode.getUserObject()).getWidget());
   }
   
   /**
    * Copies the filtered properties to system clipboard recursively starting
    * from the selected node.
    */
   public void copyPropertiesToClipboard()
   {
      Widget<?> rootWidget = null;
      
      DefaultMutableTreeNode selectedNode = 
            (DefaultMutableTreeNode) getLastSelectedPathComponent();
      
      if (selectedNode == null)
      {
         // if nothing selected use the root node
         selectedNode = (DefaultMutableTreeNode) getModel().getRoot();
      }
      
      if (selectedNode != null && selectedNode.getUserObject() instanceof WidgetWrapper)
      {
         WidgetWrapper ww = (WidgetWrapper) selectedNode.getUserObject();
         rootWidget = ww.getWidget();
      }
      
      if (rootWidget == null)
      {
         return;
      }
      
      // get the props
      StringBuilder sb = new StringBuilder();
      dumpProperties(Collections.<Widget<?>>singletonList(rootWidget), sb);
      
      // copy the resulting text to clipboard
      StringSelection stringSelection = new StringSelection(sb.toString());
      Clipboard clipboard = Toolkit.getDefaultToolkit().getSystemClipboard();
      clipboard.setContents(stringSelection, null);
   }
   
   /**
    * Recursively dumps widget properties to the passed-in string builder.
    * 
    * @param   widgets
    *          The widgets from where (including) to start dumping the properties.
    * @param   sb
    *          A valid string builder reference.
    */
   private void dumpProperties(List<Widget<?>> widgets, StringBuilder sb)
   {
      if (widgets == null || widgets.isEmpty())
      {
         return;
      }
      
      for (Widget<?> widget : widgets) 
      {
         // dump the qualifier
         sb.append("== ").append(buildWidgetQualifier(widget)).append(" ==").append('\n');
         
         // dump props
         Map<String, Object> props = getWidgetProperties(widget, inclusions);
         
         // calculate the values column position so that the values are aligned
         int maxKey = 0;
         for (String key : props.keySet())            
         {
            if (maxKey < key.length())
            {
               maxKey = key.length();
            }
         }
         
         for (Map.Entry<String, Object> prop : props.entrySet())
         {
            String key = prop.getKey();
            sb.append(key);
            
            // align the value according to the max key length
            int valSpaces = maxKey - key.length();
            if (valSpaces < 1)
            {
               valSpaces = 1;
            }
            for (int i = 0; i < valSpaces; i++)
            {
               sb.append(' ');
            }
            
            sb.append(prop.getValue()).append('\n');
         }
         
         sb.append('\n');
         
         if (widget instanceof Container<?>)
         {
            List<Widget<?>> chs = Arrays.<Widget<?>>asList(((Container<?>) widget).widgets());
            
            // recurse to the container 
            dumpProperties(chs, sb);
         } 
      }
   }
   
   /**
    * Build widget qualifier uniquely identifying its position
    * in the object tree. 
    * 
    * @param   widget
    *          A valid widget reference.
    * 
    * @return  The qualifier string.
    */
   private String buildWidgetQualifier(Widget<?> widget)
   {
      StringBuilder sb = new StringBuilder();
      while(widget != null)
      {
         if (sb.length() > 0)
         {
            sb.insert(0, "/");
         }
         sb.insert(0, formatWidget(widget));
         widget = widget.parent();
      }
      
      return sb.toString();
   }

   /**
	 * The worker method responsible for building the tree nodes in the tree component.
	 * 
	 * @param   container
	 *          The root container.
	 * 
	 * @return  The root tree node.
	 */
	private DefaultMutableTreeNode buildTreeModel(Container<?> container) 
	{
		DefaultMutableTreeNode root = new DefaultMutableTreeNode(new WidgetWrapper(container));

		if (container.widgets() == null)
		{
			return root;
		}
		
		for (Widget<?> widget : container.widgets()) 
		{
			if (widget instanceof Container<?>) 
			{
				root.add(buildTreeModel((Container<?>) widget));
			} else 
			{
				root.add(new DefaultMutableTreeNode(new WidgetWrapper(widget)));
			}
		}

		return root;
	}

	/**
	 * Sorts the tree nodes.
	 * 
	 * @param   node
	 *          The tree node where to start the sort.
	 *          
	 * @return  The root node of the sorted tree.
	 */
	private DefaultMutableTreeNode sortNodes(DefaultMutableTreeNode node) 
	{
		for (int i = 0; i < node.getChildCount() - 1; i++) 
		{
			DefaultMutableTreeNode child = (DefaultMutableTreeNode) node.getChildAt(i);
			String nt = child.getUserObject().toString();

			for (int j = i + 1; j <= node.getChildCount() - 1; j++) 
			{
				DefaultMutableTreeNode prevNode = (DefaultMutableTreeNode) node.getChildAt(j);
				String np = prevNode.getUserObject().toString();

				if (nt.compareToIgnoreCase(np) > 0) 
				{
					node.insert(child, j);
					node.insert(prevNode, i);
				}
			}
			if (child.getChildCount() > 0) 
			{
				sortNodes(child);
			}
		}

		return node;
	}

   /**
    * Returns the list of expanded widgets in the tree.
    * 
    * @return  List of widgets.
    */
	private List<Widget<?>> getExpandedWidgets()
   {
      List<Widget<?>> expandedWidgets = new ArrayList<>();
      Enumeration<TreePath> expandedPaths = 
            getExpandedDescendants(new TreePath(getModel().getRoot()));
      
      if (expandedPaths == null)
      {
         return expandedWidgets;
      }

      // iterate per each expanded path
      while (expandedPaths.hasMoreElements())
      {
         // get node for path
         Object node = expandedPaths.nextElement().getLastPathComponent();

         // store widget from node
         if (node instanceof DefaultMutableTreeNode
               && ((DefaultMutableTreeNode) node).getUserObject() instanceof WidgetWrapper)
         {
            expandedWidgets.add(((WidgetWrapper) ((DefaultMutableTreeNode) node).getUserObject())
                  .getWidget());
         }
      }

      return expandedWidgets;
   }

   /**
    * Expand the supplied tree paths.
    * 
    * @param   paths
    *          List of tree paths.
    */
	private void expandRows(List<TreePath> paths)
   {
      for (TreePath path : paths)
      {
         expandPath(path);
      }
   }

	/**
    * Finds tree paths for the supplied list of widgets.
    * 
    * @param   subjects
    *          List of {@link Widget} references.
    *          
    * @return  List of {@link TreePath} references.
    */
   private List<TreePath> findNodes(List<Widget<?>> subjects)
   {
      List<TreePath> output = new ArrayList<>();
      findNodes(getModel(), getModel().getRoot(), subjects, output);
      return output;
   }

   /**
    * Recursively finds tree paths for the supplied list of widgets.
    * 
    * @param   model
    *          {@link TreeModel} reference.
    * @param   node
    *          The root node.
    * @param   subjects
    *          List of {@link Widget} references.
    * @param   output
    *          A valid instance of List of {@link TreePath} references where to store the result. 
    */
   private void findNodes(TreeModel model, Object node, List<Widget<?>> subjects, 
                          List<TreePath> output)
   {
      if (node instanceof DefaultMutableTreeNode
            && ((DefaultMutableTreeNode) node).getUserObject() instanceof WidgetWrapper)
      {
         Widget<?> widget = ((WidgetWrapper) ((DefaultMutableTreeNode) node).getUserObject())
               .getWidget();
         
         for (Widget<?> subject : subjects)
         {
            if (widget == subject)
            {
               TreeNode[] result = ((DefaultMutableTreeNode) node).getPath();

               if (result != null)
               {
                  output.add(new TreePath(result));
               }

               if (subjects.size() == output.size())
               {
                  return;
               }
            }
         }
      }

      for (int i = 0; i < model.getChildCount(node); i++)
      {
         findNodes(model, model.getChild(node, i), subjects, output);

         if (subjects.size() == output.size())
         {
            return;
         }
      }
   }
   
   /**
    * Builds map of properties to display in the property table.
    * The map is sorted by property keys in ascending order.
    * <p>
    * The caller can pass a collection of inclusion patterns. An
    * inclusion pattern is a string that may contain a wildcard character
    * at the beginning or end of the string.
    * Properties that match any of the inclusion patterns are returned. 
    * 
    * @param   widget
    *          The widget instance which will be introspected for properties.
    * @param   inclusions
    *          The collection of inclusion patterns. May be null or empty, 
    *          in which case all properties are returned.
    * 
    * @return  map of string-object entries.
    */
   private SortedMap<String, Object> getWidgetProperties(Widget<?> widget, 
                                                         Iterable<String> inclusions)
   {
      SortedMap<String, Object> result = new TreeMap<>();
      
      // fields of the widget itself
      getAllFields(result, null, widget);

      // fields of the associated config
      WidgetConfig config = widget.config();
      if (config != null)
      {
         getAllFields(result, "config", widget.config());
      }
      
      // hashcode as object identity of the widget
      String hashcode = Integer.toHexString(System.identityHashCode(widget));
      result.put("hashcode", hashcode);
      
      result = filterProperties(result, inclusions);
      
      return result;
   }
   
   /**
    * Gathers all <code>ref</code>'s fields and puts them into <code>result</code>.
    * Field name is used the key into the map. If <code>namespace</code> is non-empty, 
    * the key is prepended with its value. 
    * 
    * @param   result
    *          The result map where the ref's field name-values are stored.
    * @param   namespace
    *          The string to be prepended with the field names to form a "compound"
    *          key in the resulting map. May be null. 
    * @param   ref
    *          The object whose fields are introspected.
    */
   private void getAllFields(Map<String, Object> result, String namespace, Object ref)
   {
      // normalize namespace
      if (namespace == null)
      {
         namespace = "";
      }
      else
      {
         namespace = namespace.trim();
         if (namespace.length() > 0)
         {
            namespace = namespace + ".";
         }
      }

      Class<?> clazz = ref.getClass();
      while (!clazz.equals(Object.class))
      {
         Field[] fields = clazz.getDeclaredFields();
         for (Field field : fields)
         {
            // skip static fields
            if (java.lang.reflect.Modifier.isStatic(field.getModifiers()))
            {
               continue;
            }

            String fieldName = field.getName();

            // skip aspectj fields
            if (fieldName.startsWith("ajc$"))
            {
               continue;
            }

            field.setAccessible(true);
            Object fieldValue = null;
            try
            {
               fieldValue = field.get(ref);
               fieldValue = formatValue(fieldValue);
            }
            catch (Exception e)
            {
               fieldValue = "<" + e.toString() + ">";
            }

            // handle the possible case when fields with the same name are declared
            // at multiple hierarchy levels, assuming the simple class names are distinct,
            // build the qualified name
            String qname = namespace + fieldName;

            if (result.containsKey(qname))
            {
               qname = namespace + clazz.getSimpleName() + "." + fieldName;
            }

            result.put(qname, fieldValue);
         }
         clazz = clazz.getSuperclass();
      }
   }

   /**
    * Filters the passed in map of properties. 
    * 
    * @param    props
    *           The property map.
    * @param    inclusions
    *           The inclusion patterns applied to the property map.
    *           
    * @return   Returns the filtered property map.
    */
   private static SortedMap<String, Object> filterProperties(SortedMap<String, Object> props, 
                                                       Iterable<String> inclusions)
   {
      SortedMap<String, Object> result = new TreeMap<>();

      for (Map.Entry<String, Object> prop : props.entrySet())
      {
         boolean include = true;
         if (inclusions != null)
         {
            for (String inclusion : inclusions)
            {
               if (inclusion == null)
               {
                  continue;
               }
               inclusion = inclusion.trim();
               if (inclusion.isEmpty())
               {
                  continue;
               }

               include = false;
               if (prop.getKey().toLowerCase().contains(inclusion.toLowerCase()))
               {
                  include = true;
                  break;
               }
            }
         }

         if (!include)
         {
            continue;
         }

         result.put(prop.getKey(), prop.getValue());
      }

      return result;
   }
   
   /**
    * Formats the supplied reference into a string.
    * 
    * @param   obj
    *          A reference, may be <code>null</code>.
    * @return  The formatted value.
    */
   private String formatValue(Object obj)
   {
      if (obj instanceof Widget<?>)
      {
         return formatWidget((Widget<?>) obj);
      }
      
      if (obj instanceof WeakReference)
      {
         obj = ((WeakReference<?>) obj).get(); 
      }
         
      if (obj instanceof Point ||
          obj instanceof NativePoint ||
          obj instanceof Dimension ||
          obj instanceof NativeDimension ||
          obj instanceof Insets ||
          obj instanceof Rectangle ||
          obj instanceof NativeRectangle ||
          obj instanceof NativeInsets ||
          obj instanceof Color)
      {
         return ToStringBuilder.reflectionToString(obj, ToStringStyle.SHORT_PREFIX_STYLE);
      }
      
      return Objects.toString(obj);
   }

   /**
    * Formats the supplied reference into a string.
    * 
    * @param   widget
    *          A valid {@link Widget} reference.
    *          
    * @return  The formatted string.
    */
   private String formatWidget(Widget<?> widget)
   {
      String className = showFullClassName ? widget.getClass().getName() : widget.getClass()
            .getSimpleName();

      String id = widget.config() == null || widget.config().id == null ? "null" : widget
            .config().id.toString();

      return String.format("%s<%s>", className, id);
   }
   
   /**
    * Wrapper class to help store {@link Widget} references 
    * in the tree component.  
    */
   public class WidgetWrapper 
   {
      /**
       * The widget reference.
       */
      private Widget<?> widget;

      /**
       * Constructor.
       * 
       * @param   widget
       *          A valid {@link Widget} reference.
       */
      public WidgetWrapper(Widget<?> widget) 
      {
         this.widget = widget;
      }

      /**
       * Returns the wrapped widget reference.
       * 
       * @return  {@link Widget} reference.
       */
      public Widget<?> getWidget() 
      {
         return widget;
      }
      
      /**
       * Builds map of properties to display in the property table.
       * The map is sorted by property keys in ascending order.
       * <p>
       * The caller can pass a collection of inclusion patterns. An
       * inclusion pattern is a string that may contain a wildcard character
       * at the beginning or end of the string.
       * Properties that match any of the inclusion patterns are returned. 
       * 
       * @param   inclusions
       *          The collection of inclusion patterns. May be null or empty, 
       *          in which case all properties are returned.
       * 
       * @return  map of string-object entries.
       */
      public SortedMap<String, Object> getProperties(Iterable<String> inclusions)
      {
         return getWidgetProperties(widget, inclusions);
      }
   
      /**
       * This override is called by the tree component to show the node label.
       * 
       * @return  String representing this widget in the tree component.
       */
      @Override
      public String toString() 
      {
         return formatWidget(widget);
      }
   }
}
