/*
** Module   : PropertyTable.java
** Abstract : The tree Widget Browser component.
**
** Copyright (c) 2014-2015, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- ---------------------------------Description----------------------------------
** 001 HC  20141031 Created initial version
** 002 HC  20141130 Added the capability to display widget config fields.
** 003 HC  20150216 Widget properties can be copied to clipboard.
*/
package com.goldencode.p2j.ui.client.widgetbrowser;

import java.awt.Component;
import java.awt.event.ActionListener;
import java.awt.event.MouseAdapter;
import java.awt.event.MouseEvent;
import java.util.*;

import javax.swing.*;
import javax.swing.table.*;

import org.apache.commons.lang.StringEscapeUtils;

/**
 * The tree Widget Browser component.
 */
@SuppressWarnings("serial")
public class PropertyTable 
extends JTable 
{
   /**
    * Column names.
    */
   private String[] tableColumns = { "Property", "Value" };

   /**
    * Table model.
    */
   private DefaultTableModel tableModel;

   /**
    * Event listener.
    */
   private ActionListener actionListener;

   /**
    * Constructor.
    */
   public PropertyTable() 
   {
      // set the tooltip delay
      ToolTipManager.sharedInstance().setDismissDelay(120000);
      
      // create table model
      tableModel = new DefaultTableModel();
      tableModel.setColumnIdentifiers(tableColumns);
      setModel(tableModel);

      addMouseListener(new MouseAdapter() 
      {
         public void mouseClicked(MouseEvent e) 
         {
            // check double click
            if (e.getClickCount() != 2) 
            {
               return;
            }

            // if clicked row is widget property, redirect to the widget node
            if (actionListener != null)
            {
               int row = PropertyTable.this.rowAtPoint(e.getPoint());
               if (row >= 0)
               {
                  actionListener.actionPerformed(
                        new Event(PropertyTable.this, Event.PROPERTY_VALUE_NAVIGATED, 
                                  PropertyTable.this.dataModel.getValueAt(row, 1)));
               }
            }
         }
      });
   }

   /**
    * The override disables cell editing.
    */
   @Override
   public boolean isCellEditable(int row, int column) 
   {
      return false;
   };
   
   /**
    * The override adds tooltips to the value cells.
    * 
    * @param   renderer  
    *          the <code>TableCellRenderer</code> to prepare
    * @param   row       
    *          the row of the cell to render, where 0 is the first row
    * @param   column    
    *          the column of the cell to render,
    *          where 0 is the first column
    *        
    * @return  the <code>Component</code> under the event location
    */
   @Override
   public Component prepareRenderer(TableCellRenderer renderer, int row, int column)
   {
      Component c = super.prepareRenderer(renderer, row, column);
      
      // tooltips only for property values
      if (column != 1)
      {
         return c;
      }
      
      if (c instanceof JComponent)
      {
         JComponent jc = (JComponent) c;
         
         String text = getValueAt(row, column).toString();
         
         // make the value multi-line
         text = StringEscapeUtils.escapeHtml(text);
         String[] parts = text.split("\\s");
         StringBuffer buf = new StringBuffer();
         buf.append("<html>");
         int lineLength = 0;
         for (String part : parts)
         {
            if (lineLength + part.length() > 80 && lineLength > 0)
            {
               lineLength = 0;
               buf.append("<br>");
            }
            
            buf.append(part);
            buf.append(" ");
            lineLength += part.length();
         }
         buf.append("</html>");     
         jc.setToolTipText(buf.toString());
      }
      return c;
   }

   /**
    * Sets the event handler.
    * 
    * @param   listener
    *          Reference to {@link ActionListener} or <code>null</code>.
    */
   public void setActionListener(ActionListener listener) 
   {
      actionListener = listener;
   }

   /**
    * Clears all property rows.
    */
   public void resetProperties() 
   {
      tableModel.setRowCount(0);
   }
   
   /**
    * Assign the properties to be shown in the table.
    * 
    * @param   properties
    *          The sorted property map.
    */
   public void setTableContent(SortedMap<String, Object> properties) 
   {
      // clear old content
      tableModel.setRowCount(0);
      List<TableRow> rows = new ArrayList<>();

      for (Map.Entry<String, Object> entry : properties.entrySet()) 
      {
         // substitute null with "<null>"
         Object value = entry.getValue();
         if (value == null)
         {
            value = "<null>";
         }
         
         // add object, toString() is called automatically
         rows.add(new TableRow(entry.getKey(), value));
      }

      // add rows to the table
      for (TableRow tableRow : rows) 
      {
         Object[] row = {tableRow.getName(), tableRow.getValue()};
         tableModel.addRow(row);
      }
   }
   
   /**
    * The instance representing one row in the table.
    */
   private static class TableRow 
   implements Comparable<TableRow> 
   {
      /**
       * The name of the property. 
       */
      private String name;

      /**
       * The value of the property.
       */
      private Object value;

      /**
       * The constructor.
       * 
       * @param   name
       *          Property name.
       * @param   value
       *          Property value.
       */
      public TableRow(String name, Object value) 
      {
         super();
         this.name = name;
         this.value = value;
      }

      /**
       * Name getter.
       * 
       * @return  Property name.
       */
      public String getName() 
      {
         return name;
      }

      /**
       * Value getter.
       * 
       * @return  Property value.
       */
      public Object getValue() 
      {
         return value;
      }

      /**
       * Comparator to compare by the property names.
       */
      @Override
      public int compareTo(TableRow o) 
      {
         return getName().compareTo(o.getName());
      }
      
   }
}
